"""
A module for working with two-dimensional planes.
"""

# built-in
from typing import Iterable, NamedTuple, Union

# internal
from svgen.cartesian.mutate import Translation
from svgen.cartesian.point import DEFAULT, NamedPoints, Point, PointManager
from svgen.cartesian.rectangle import Rectangle


class RectangleGrid(NamedTuple):
    """A class for locating points within a rectangle."""

    rect: Rectangle
    columns: int
    rows: int

    @property
    def x_scale(self) -> float:
        """Get the scalar value for 'x' coordinates."""
        return self.rect.width / float(self.columns)

    @property
    def y_scale(self) -> float:
        """Get the scalar value for 'y' coordinates."""
        return self.rect.height / float(self.rows)

    def __call__(self, x: float, y: float) -> Point:
        """Get the true location of a point within this grid."""

        # Ensure the coordinates are within the sub-divisions.
        assert x < self.columns
        assert y < self.rows
        return self.rect.origin.translate(x * self.x_scale, y * self.y_scale)

    @property
    def points(self) -> Iterable[Point]:
        """Iterate over points in the grid."""
        for row in range(self.rows):
            for col in range(self.columns):
                yield self(col, row)

    def box(self, x: float, y: float) -> Rectangle:
        """Get a rectangle object from provided coordinates."""
        return Rectangle.create(self.x_scale, self.y_scale, self(x, y))

    @property
    def boxes(self) -> Iterable[Rectangle]:
        """Iterate over boxes belonging to this grid."""
        for row in range(self.rows):
            for col in range(self.columns):
                yield self.box(col, row)

    def adjust(
        self, columns: int = None, rows: int = None, rect: Rectangle = None
    ) -> "RectangleGrid":
        """Create a new grid from new parameters."""
        if columns is None:
            columns = self.columns
        if rows is None:
            rows = self.rows
        if rect is None:
            rect = self.rect
        return RectangleGrid(rect, columns, rows)

    def translate(
        self, dx: Union[Translation, float], *args, **kwargs
    ) -> "RectangleGrid":
        """Translate this rectangle grid."""

        return RectangleGrid(
            self.rect.translate(dx, *args, **kwargs), self.columns, self.rows
        )


class Plane(PointManager):
    """A class for managing geometric entities."""

    def __init__(
        self, origin: Point = DEFAULT, points: NamedPoints = None
    ) -> None:
        """Initialize this plane."""

        super().__init__(points)
        self.add_point("origin", origin)

    @property
    def origin(self) -> Point:
        """Get the origin of this plane."""
        return self.get_point("origin")



"""
A module for project-specific task registration.
"""

# built-in
from pathlib import Path
from typing import Dict

# third-party
from vcorelib.task import Phony
from vcorelib.task.manager import TaskManager
from vcorelib.task.subprocess.run import is_windows


def register(
    manager: TaskManager,
    project: str,
    cwd: Path,
    substitutions: Dict[str, str],
) -> bool:
    """Register project tasks to the manager."""

    # Don't run yamllint on Windows because it will fail on newlines.
    manager.register(
        Phony("yaml"),
        [] if is_windows() else ["yaml-lint-local", "yaml-lint-manifest.yaml"],
    )
    del project
    del cwd
    del substitutions
    return True



"""
svgen - Common interfaces and assets for SVG elements.
"""

# built-in
from io import StringIO
import os
from typing import Dict, List, TextIO, cast
from xml.etree import ElementTree as et

# internal
from svgen.attribute import Attribute, attributes
from svgen.attribute.style import Style

INDENT: int = 2


class Element:
    """A basic interface for an SVG element."""

    def __init__(
        self,
        tag: str = None,
        text: str = "",
        attrib: List[Attribute] = None,
        children: List["Element"] = None,
        **extra,
    ) -> None:
        """Construct a new SVG element."""

        if tag is None:
            tag = type(self).__name__
            tag = tag[0].lower() + tag[1:]
        if attrib is None:
            attrib = []
        if children is None:
            children = []

        self.tag = tag
        self.text = text

        self.attributes: Dict[str, Attribute] = {}
        for attr in attrib + attributes(extra):
            self.add_attribute(attr)

        self.children: List[Element] = children

    @property
    def xml(self) -> et.Element:
        """Get this element as an xml element."""

        elem = et.Element(
            self.tag, {x.key: x.value for x in self.attributes.values()}
        )
        elem.text = self.text

        # Add children.
        elem.extend(child.xml for child in self.children)
        return elem

    @property
    def style(self) -> Style:
        """Get the style attribute for this element."""

        if "style" not in self.attributes:
            self.add_attribute(Style())
        return cast(Style, self.attributes["style"])

    def add_attribute(self, attr: Attribute) -> "Element":
        """Add an attribute to this element."""
        assert (
            attr.key not in self.attributes
        ), f"This '{self.tag}' element already has attribute '{attr.key}'!"
        self.attributes[attr.key] = attr
        return self

    def closing(self, indent: int = 0) -> str:
        """Create a string to close this element."""

        if not self.text and not self.children:
            return " />"

        indent_str = " " * (indent * INDENT)
        return f"{indent_str}</{self.tag}>"

    def encode(
        self,
        output: TextIO,
        quote: str = '"',
        indent: int = 0,
        newlines: bool = True,
    ) -> None:
        """Encode this element to a string stream."""

        # Indent will matter if we want lines.
        indent_str = " " * (indent * INDENT)
        if newlines:
            output.write(indent_str)

        attr_strs = [x.encode(quote) for x in self.attributes.values()]
        attrs = " ".join(x for x in attr_strs if x)
        output.write(f"<{self.tag}")
        if attrs:
            output.write(f" {attrs}")

        if self.text or self.children:
            # Close the opening tag.
            output.write(">")
            if newlines:
                output.write(os.linesep)

            # Write content, if any.
            if newlines and self.text:
                output.write(indent_str + " " * INDENT)
            if self.text:
                output.write(self.text)
                if newlines:
                    output.write(os.linesep)

            # Write children.
            for child in self.children:
                if child:
                    child.encode(output, quote, indent + 1, newlines)

        # Close this element tag.
        output.write(self.closing(indent))
        if newlines:
            output.write(os.linesep)

    def encode_str(
        self,
        quote: str = '"',
        indent: int = 0,
        newlines: bool = True,
    ) -> str:
        """Encode this element to a string."""

        with StringIO() as output:
            self.encode(output, quote, indent, newlines)
            return output.getvalue()



"""
A simpler setuptools-based package definition.
"""

# built-in
from contextlib import contextmanager
import os
import shutil
import tempfile
from typing import Any, Dict, Iterator, List, Set, Union, cast

# third-party
import setuptools

SELF = "setuptools-wrapper"


@contextmanager
def inject_self(
    working_dir: str,
    curr_pkg_slug: str,
    pkg: str = SELF,
    force_copy: bool = False,
) -> Iterator[None]:
    """
    Copy this entire package into the caller's source distribution. This is
    the only way to avoid a pointless requirement to already have this package
    installed to install anything else, and also generally not requiring
    an explicit install of this package except for command-line use.
    """

    added = False
    pkg = pkg.replace("-", "_")

    # inject sources into a package with a different name, otherwise
    # installation becomes very broken and messed up
    to_create = os.path.join(working_dir, f"{curr_pkg_slug}_bootstrap")

    try:
        # do nothing if we are building ourselves (but allow forcing this)
        if (force_copy or pkg not in curr_pkg_slug) and not os.path.isdir(
            to_create
        ):
            os.mkdir(to_create)

            # copy our sources into their package
            to_copy = [
                "__init__.py",
                "setup.py",
                "py.typed",
            ]
            vmklib_dir = os.path.dirname(__file__)
            for fname in to_copy:
                dest = os.path.join(to_create, fname)
                src = os.path.join(vmklib_dir, fname)
                if not os.path.isfile(dest) and os.path.isfile(src):
                    shutil.copyfile(src, dest)

            added = True

        yield
    finally:
        if added:
            shutil.rmtree(to_create)


def get_long_description(desc_filename: str = "README.md") -> str:
    """Get a package's long-description data from a file."""

    try:
        with open(desc_filename, "r", encoding="utf-8") as desc_file:
            long_description = desc_file.read()
        return long_description
    except FileNotFoundError:
        return ""


def default_requirements_file(directory: str) -> str:
    """Default location to look for the requirements file."""

    return os.path.join(directory, "requirements.txt")


def get_requirements(reqs_filename: str) -> Set[str]:
    """Get a package's requirements based on its requirements file."""

    try:
        with open(reqs_filename, "r", encoding="utf-8") as reqs_file:
            reqs: List[str] = reqs_file.read().strip().split()
        return set(reqs)
    except FileNotFoundError:
        return set()


def get_data_files(pkg_name: str, data_dir: str = "data") -> List[str]:
    """
    Get the non-code sources under a package directory's data directory.
    """

    data_files = []
    for root, _, files in os.walk(os.path.join(pkg_name, data_dir)):
        for fname in files:
            rel_name = os.path.join(root, fname).replace(pkg_name + os.sep, "")
            data_files.append(rel_name)

    return data_files


# pylint: disable=too-many-arguments
def setup(
    pkg_info: Dict[str, Any],
    author_info: Dict[str, str],
    url_override: str = None,
    classifiers_override: List[str] = None,
    requirements: Set[str] = None,
    **kwargs,
) -> None:
    """
    Build a 'setuptools.setup' call with sane defaults and making assumptions
    about certain aspects of a package's structure.
    """

    defaults: Dict[str, Union[str, List[str], Set[str]]] = {
        "url_override": (
            f"https://github.com/{author_info['username']}/"
            f"{pkg_info['name']}"
        ),
        "classifiers_override": [
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        "requirements": set(),
    }

    # Resolve defaults if necessary.
    url_override = cast(
        str, defaults["url_override"] if url_override is None else url_override
    )
    classifiers_override = cast(
        List[str],
        defaults["classifiers_override"]
        if classifiers_override is None
        else classifiers_override,
    )
    requirements = cast(
        Set[str],
        defaults["requirements"] if requirements is None else requirements,
    )

    for version in pkg_info.get("versions", []):
        classifiers_override.append(
            f"Programming Language :: Python :: {version}"
        )

    # Find requirements files inside the package's root directory.
    req_files = [default_requirements_file(pkg_info["slug"])]
    for req_file in req_files:
        requirements |= get_requirements(req_file)

    with tempfile.TemporaryDirectory() as temp_dir:
        working_dir = temp_dir
        dir_contents = os.listdir(os.getcwd())
        if pkg_info["slug"] in dir_contents:
            working_dir = os.getcwd()

        with inject_self(
            working_dir,
            pkg_info["slug"],
            force_copy=pkg_info.get("force_copy", False),
        ):
            setuptools.setup(
                name=pkg_info["name"],
                version=pkg_info["version"],
                author=author_info["name"],
                author_email=author_info["email"],
                description=pkg_info["description"],
                long_description=get_long_description(),
                long_description_content_type="text/markdown",
                url=url_override,
                packages=setuptools.find_packages(
                    exclude=["tests", "tests.*"]
                ),
                classifiers=classifiers_override,
                python_requires=f">={pkg_info.get('versions', ['3.6'])[0]}",
                install_requires=list(requirements),
                package_data={
                    pkg_info["slug"]: (
                        get_data_files(pkg_info["slug"])
                        + ["py.typed", "*.pyi", "*.txt"]
                    ),
                    f"{pkg_info['slug']}_bootstrap": ["py.typed"],
                    "": ["*.pyi"],
                },
                **kwargs,
            )



"""
svgen - A module for the 'viewBox' attribute.
"""

# built-in
from typing import NamedTuple, Union

# internal
from svgen.attribute import Attribute
from svgen.cartesian import Point
from svgen.cartesian.mutate import Translation
from svgen.cartesian.plane import Plane, RectangleGrid
from svgen.cartesian.rectangle import Dimensions, Rectangle


class ViewBoxData(NamedTuple):
    """Primitive data for a viewBox attribute."""

    min_x: int
    min_y: int
    width: int
    height: int

    def translate(self, dx: int, dy: int) -> "ViewBoxData":
        """Translate this viewBox."""
        return ViewBoxData(
            self.min_x + dx, self.min_y + dy, self.width, self.height
        )

    @property
    def origin(self) -> Point:
        """Get the origin for this viewBox."""
        return Point(float(self.min_x), float(self.min_y))

    @property
    def center(self) -> Point:
        """Get the center point for this viewBox."""
        return Point(
            self.min_x + (float(self.width) / 2.0),
            self.min_y + (float(self.height) / 2.0),
            True,
        )

    @staticmethod
    def from_dict(data: dict) -> "ViewBoxData":
        """Create viewBox data from a dictionary."""
        return ViewBoxData(
            data.get("min_x", 0),
            data.get("min_y", 0),
            data["width"],
            data["height"],
        )


class ViewBox(Attribute):
    """An interface for viewBox attributes."""

    def __init__(
        self, min_x: int, min_y: int, width: int, height: int
    ) -> None:
        """Construct a new viewBox."""

        self.data = ViewBoxData(min_x, min_y, width, height)
        self.plane = Plane(self.data.origin)
        self.box = Rectangle(self.dimensions, self.origin)
        self.grid = RectangleGrid(self.box, width, height)

    def new_grid(
        self, rect: Rectangle = None, columns: int = None, rows: int = None
    ) -> RectangleGrid:
        """Create a grid from this viewBox."""
        return self.grid.adjust(columns, rows, rect)

    def translate(
        self, dx: Union[Translation, float], *args, **kwargs
    ) -> None:
        """Apply a translation to this viewBox."""

        move = Translation.normalize(dx, *args, **kwargs)

        # Update instance attributes.
        self.data = self.data.translate(int(move.dx), int(move.dy))
        self.box = self.box.translate(move)
        self.plane.translate(move)
        self.grid = self.grid.translate(move)

    @property
    def dimensions(self) -> Dimensions:
        """Get this viewBox's dimensions."""
        return Dimensions(float(self.data.width), float(self.data.height))

    @property
    def origin(self) -> Point:
        """Get this viewBox's origin."""
        return self.plane.origin

    @property
    def key(self) -> str:
        """Get the string key for this attribute."""
        return "viewBox"

    @property
    def value(self) -> str:
        """Get the string value for this attribute."""
        return (
            f"{self.data.min_x} {self.data.min_y} "
            f"{self.data.width} {self.data.height}"
        )

    @property
    def center(self) -> Point:
        """Get the center point for this viewBox."""
        return self.data.center

    @staticmethod
    def decode(key: str, value: str) -> Attribute:
        """Create this attribute from a string."""

        assert key == "viewBox"
        args = [int(x) for x in value.split()]
        assert len(args) == 4
        return ViewBox(*args)

    @staticmethod
    def from_dict(data: dict) -> "ViewBox":
        """Create a viewBox from dictionary data."""
        return ViewBox(*ViewBoxData.from_dict(data))



# =====================================
# generator=datazen
# version=2.1.0
# hash=c899afd4b2308889f0dc1260eede4d46
# =====================================
"""
svgen - Package's default entry-point.
"""

# built-in
import sys

# internal
from svgen.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



# =====================================
# generator=datazen
# version=2.1.0
# hash=21abdf006353a5984d0a37e7e2854553
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A simple interface to setuptools's setup function."
PKG_NAME = "setuptools-wrapper"
VERSION = "0.1.2"



# =====================================
# generator=datazen
# version=2.1.0
# hash=6bb3085a425c3a2b745ee35ac0fe5b3e
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A tool for working with scalable vector graphics."
PKG_NAME = "svgen"
VERSION = "0.4.0"
